__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

import chess_board
import san

class ChessPlayer:
    
    __name = '(unnamed)'
    __type = None
    
    __game = None
    
    __readyToMove = False
    
    def __init__(self, name):
        self.setName(name)
        
    def setName(self, name):
        self.__name = name
        
    def getName(self):
        return self.__name
    
    def setType(self, type):
        self.__type = type
        
    def getType(self):
        return self.__type
        
    def onPieceMoved(self, piece, start, end):
        """
        """
        pass
    
    def onPlayerMoved(self, player, sanMove): # FIXME: Replace SAN move with a class containing the move in various forms
        """
        """
        pass
    
    def onGameEnded(self, winningPlayer = None):
        """
        """
        pass
    
    # Public methods
    
    def readyToMove(self):
        """
        """
        pass
    
    def _setReadyToMove(self, readyToMove):
        self.__readyToMove = readyToMove
        if readyToMove is True:
            self.readyToMove()
    
    def readyToMove(self):
        """
        """
        return self.__readyToMove
    
    def canMove(self, move, promotionType = chess_board.QUEEN):
        """
        """
        return self.__game.canMove(self, move, promotionType)

    def move(self, move):
        """
        """
        self.__game.move(self, move)

    def setGame(self, game):
        """
        """
        self.__game = game

class ChessGameBoard(chess_board.ChessBoard):
    """
    """
    
    # Reference to the game
    __game = None
    
    def __init__(self, game):
        """
        """
        self.__game = game
        chess_board.ChessBoard.__init__(self)

    def onPieceMoved(self, piece, start, end):
        """Called by chess_board.ChessBoard"""
        self.__game._onPieceMoved(piece, start, end)

class ChessGameSANConverter(san.SANConverter):
    """
    """
        
    __colourToSAN = {chess_board.WHITE: san.SANConverter.WHITE,
                     chess_board.BLACK: san.SANConverter.BLACK}
    __sanToColour = {}
    for (a, b) in __colourToSAN.iteritems():
        __sanToColour[b] = a
        
    __typeToSAN = {chess_board.PAWN:   san.SANConverter.PAWN,
                   chess_board.KNIGHT: san.SANConverter.KNIGHT,
                   chess_board.BISHOP: san.SANConverter.BISHOP,
                   chess_board.ROOK:   san.SANConverter.ROOK,
                   chess_board.QUEEN:  san.SANConverter.QUEEN,
                   chess_board.KING:   san.SANConverter.KING}
    __sanToType = {}
    for (a, b) in __typeToSAN.iteritems():
        __sanToType[b] = a
        
    __resultToSAN = {chess_board.MOVE_RESULT_ILLEGAL: False,
                     chess_board.MOVE_RESULT_ALLOWED: True,
                     chess_board.MOVE_RESULT_OPPONENT_CHECK: san.SANConverter.CHECK,
                     chess_board.MOVE_RESULT_OPPONENT_CHECKMATE: san.SANConverter.CHECKMATE}
        
    __board = None
        
    def __init__(self, board):
        self.__board = board
    
    def decodeSAN(self, colour, move):
        (start, end, result, promotionType) = san.SANConverter.decodeSAN(self, self.__colourToSAN[colour], move)
        return (start, end, self.__sanToType[promotionType])
    
    def encodeSAN(self, start, end, promotionType):
        return san.SANConverter.encodeSAN(self, start, end, self.__typeToSAN[promotionType])

    def getPiece(self, file, rank):
        """Called by san.SANConverter"""
        piece = self.__board.getPiece((file, rank))
        if piece is None:
            return None
        return (self.__colourToSAN[piece.getColour()], self.__typeToSAN[piece.getType()])
    
    def testMove(self, colour, start, end, promotionType, allowSuicide = False):
        """Called by san.SANConverter"""
        moveResult = self.__board.testMove(self.__sanToColour[colour], ((start, end)), self.__sanToType[promotionType], allowSuicide)
        
        return self.__resultToSAN[moveResult]
    
class ChessGameCANConverter:
    """
    """
    
    __board = None
    
    __fileToCAN = {0:'a', 1:'b', 2:'c', 3:'d', 4:'e', 5:'f', 6:'g', 7:'h'}
    __rankToCAN = {0:'1', 1:'2', 2:'3', 3:'4', 4:'5', 5:'6', 6:'7', 7:'8'}
    __canToFile = {'a':0, 'b':1, 'c':2, 'd':3, 'e':4, 'f':5, 'g':6, 'h':7}
    __canToRank = {'1':0, '2':1, '3':2, '4':3, '5':4, '6':5, '7':6, '8':7}
    
    __typeToCAN = {chess_board.PAWN:   'p',
                   chess_board.KNIGHT: 'n',
                   chess_board.BISHOP: 'b',
                   chess_board.ROOK:   'r',
                   chess_board.QUEEN:  'q',
                   chess_board.KING:   'k'}
    __canToType = {}
    for (a, b) in __typeToCAN.iteritems():
        __canToType[b] = a
    
    def __init__(self, board):
        """
        """
        self.__board = board
        
    def moveIsCAN(self, move):
        """Test if a move is in CAN format.
        
        'move' is the move to test (string).
        
        Return True if this move is in CAN format otherwise False.
        """
        if len(move) < 4:
            return False
        elif len(move) > 5:
            return False
        
        # Check move is in the form file, rank, file, rank
        files = 'abcdefgh'
        ranks = '12345678'
        if files.find(move[0]) < 0 or ranks.find(move[1]) < 0 or \
           files.find(move[2]) < 0 or ranks.find(move[3]) < 0:
            return False
        
        if len(move) == 5:
            if 'nbrq'.find(move[4]) < 0:
                return False
        return True
    
    def encodeCAN(self, start, end, promotionType):
        """Encode a move in Co-ordinate Algebraic Notation (CAN).
        
        'start' is the location of the piece being moved (tuple (file,rank)).
        'end' is the location of the piece is being moved to (tuple (file,rank)).
        'promotionType' is the piece type pawns are promoted to.
        
        The move is not checked if it is valid.
        """
        move = self.__fileToCAN[start[0]] + self.__rankToCAN[start[1]] + self.__fileToCAN[end[0]] + self.__rankToCAN[end[1]]
        
        # Provide promotion type if required
        piece = self.__board.getPiece(start)
        if piece is None:
            return move
        
        # If the moved piece is a pawn and it has moved to an edge of the board then promote it.
        # Note: Since pawns can't move backwards we don't have to check colour.
        if piece.getType() is chess_board.PAWN and (end[1] == 0 or end[1] == 7):
            move += self.__typeToCAN[promotionType]
        
        return move

    def decodeCAN(self, move):
        """Decode a move in Co-ordinate Algebraic Notation (CAN).
        
        'move' is the CAN move to decode (string).
        
        Returns a tuple containing (start, end, promotionType).
        If the move is invalid then a TypeError exception is raised.
        """
        if len(move) < 4:
            raise TypeError('Move ' + repr(move) + ' is too short for CAN')
        elif len(move) > 5:
            raise TypeError('Move ' + repr(move) + ' is too long for CAN')
            
        try:
            start = (self.__canToFile[move[0]], self.__canToRank[move[1]])
            end = (self.__canToFile[move[2]], self.__canToRank[move[3]])
        except KeyError:
            raise TypeError('Move ' + repr(move) + ' contains invalid co-ordinates')
        
        if len(move) == 4:
            return (start, end, chess_board.QUEEN)
        
        try:
            promotionType = self.__canToType[move[4]]
        except KeyError:
            raise TypeError('Move ' + repr(move) + ' contains invalid promotion type')
        
        return (start, end, promotionType)

class ChessGame:
    """
    """
    
    # The players and spectators in the game
    __players = None
    __whitePlayer = None
    __blackPlayer = None
    __spectators = None
    
    # The board to move on
    __board = None
    
    # SAN and CAN en/decoders
    __sanConverter = None
    __canConverter = None
    
    # The game state (started and player to move)
    __started = False
    __currentPlayer = None

    # Flag to show if calling a player and the queued up moves
    __notifyingPlayer = False
    __queuedMoves = None
    
    __sanMoves = None
    
    def __init__(self):
        """Game constructor"""
        self.__players = []
        self.__spectators = []
        self.__board = ChessGameBoard(self)
        self.__sanConverter = ChessGameSANConverter(self.__board)
        self.__canConverter = ChessGameCANConverter(self.__board)
        self.__sanMoves = []

    def setWhite(self, player):
        """Set the white player in the game.
        
        'player' is the player to use as white.
        
        If the game has started or there is a white player an exception is thrown.
        """
        assert(self.__started is False)
        assert(self.__whitePlayer is None)
        self.__whitePlayer = player
        self.__connectPlayer(player)
        
    def getWhite(self):
        return self.__whitePlayer
    
    def setBlack(self, player):
        """Set the black player in the game.
        
        'player' is the player to use as black.
        
        If the game has started or there is a black player an exception is thrown.
        """
        assert(self.__started is False)
        assert(self.__blackPlayer is None)
        self.__blackPlayer = player
        self.__connectPlayer(player)
        
    def getBlack(self):
        return self.__blackPlayer
    
    def addSpectator(self, player):
        """Add a spectator to the game.
        
        'player' is the player spectating.
        
        This can be called after the game has started.
        """
        self.__spectators.append(player)
        self.__connectPlayer(player)
        
    def start(self, moves = []):
        """Start the game.
        
        'moves' is a list of moves to start with.
        
        If there is no white or black player then an exception is raised.
        """
        assert(self.__whitePlayer is not None and self.__blackPlayer is not None)
        #print self.__board
        
        # Set initial state
        self.__queuedMoves = []
        self.__currentPlayer = self.__whitePlayer
        
        # Load starting moves
        try:
            for move in moves:
                self.move(self.__currentPlayer, move)
        except san.SANError, e:
            print e
            
        self.__started = True
        
        # Get the next player to move
        self.__currentPlayer._setReadyToMove(True)
        
    def getSquareOwner(self, coord):
        """TODO
        """
        piece = self.__board.getPiece(coord)
        if piece is None:
            return None
        
        colour = piece.getColour()
        if colour is chess_board.WHITE:
            return self.__whitePlayer
        elif colour is chess_board.BLACK:
            return self.__blackPlayer
        else:
            return None
        
    def canMove(self, player, move, promotionType):
        """Test if a player can move.
        
        'player' is the player making the move.
        'move' is the move to make (TODO).
        'promotionType' is the piece type to promote pawns to. FIXME: Make this a property of the player
        
        Return True if can move, otherwise False.
        """
        if player is not self.__currentPlayer:
            return False
        
        if player is self.__whitePlayer:
            colour = chess_board.WHITE
        elif player is self.__blackPlayer:
            colour = chess_board.BLACK
        else:
            assert(False)

        # FIXME: Convert move from SAN etc
        
        moveResult = self.__board.testMove(colour, move, promotionType = promotionType)
         
        return moveResult is not chess_board.MOVE_RESULT_ILLEGAL
    
    def move(self, player, move):
        """Get a player to make a move.
        
        'player' is the player making the move.
        'move' is the move to make. It can be of the form:
               A coordinate move in the form ((file0, rank0), (file1, rank1), promotionType) ((int, int), (int, int), chess_board.PIECE_TYPE) or
               A SAN move (string).
        """
        self.__queuedMoves.append((player, move))
        
        # Don't process if still finishing the last move
        if self.__notifyingPlayer:
            return
        
        while True:
            try:
                (movingPlayer, move) = self.__queuedMoves.pop(0)
            except IndexError:
                return
            
            if movingPlayer is not self.__currentPlayer:
                print 'Player moved out of turn'
                continue
            
            if self.__currentPlayer is self.__whitePlayer:
                nextPlayer = self.__blackPlayer
                colour = chess_board.WHITE
            else:
                nextPlayer = self.__whitePlayer
                colour = chess_board.BLACK
                
            # If move is SAN process it as such
            if type(move) is str:
                if self.__canConverter.moveIsCAN(move):
                    (start, end, promotionType) = self.__canConverter.decodeCAN(move)
                else:
                    (start, end, promotionType) = self.__sanConverter.decodeSAN(colour, move)
            elif type(move) is tuple and len(move) == 3:
                start = move[0]
                end = move[1]
                promotionType = move[2]
            else:
                raise TypeError() # FIXME

            # FIXME: Don't apply if the move is invalid
            sanMove = self.__sanConverter.encodeSAN(start, end, promotionType)
            canMove = self.__canConverter.encodeCAN(start, end, promotionType)
            moveResult = self.__board.movePiece(colour, (start, end), promotionType)
            
            if moveResult is chess_board.MOVE_RESULT_ILLEGAL:
                print 'Illegal move: ' + str(move)
                continue
            
            self.__sanMoves.append(sanMove)

            # This player has now moved
            self.__currentPlayer._setReadyToMove(False)
            
            #print str((start, end)) + ' -> ' + str(sanMove)
            #print self.__board
            
            #if type(move) is str:
            #    assert(move == sanMove)

            self.__notifyingPlayer = True
            
            # Inform other players of the result
            for player in self.__players:
                if self.__started is True and player is self.__currentPlayer:
                    continue
                player.onPlayerMoved(self.__currentPlayer, canMove) # TEMP: Use CAN for the AI
            
            # Notify the next player they can move
            self.__currentPlayer = nextPlayer
            if self.__started is True:
                nextPlayer._setReadyToMove(True)
            
            self.__notifyingPlayer = False
            
    def getSANMoves(self):
        """
        """
        return self.__sanMoves[:]
    
    def end(self):
        """End the game"""
        # Inform players
        for player in self.__players:
            player.onGameEnded()

    # Private methods:

    def __connectPlayer(self, player):
        """Add a player into the game.
        
        'player' is the player to add.
        
        The player will be notified of the current state of the board.
        """
        self.__players.append(player)
        player.setGame(self)
        
        # Notify the player of the current state
        # FIXME: Make the board iteratable...
        for file in [0, 1, 2, 3, 4, 5, 6, 7]:
            for rank in [0, 1, 2, 3, 4, 5, 6, 7]:
                coord = (file, rank)
                piece = self.__board.getPiece(coord)
                if piece is None:
                    continue

                # These are moves from nowhere to their current location
                player.onPieceMoved(piece, None, coord)
                
    def _onPieceMoved(self, piece, start, end):
        """Called by the chess board"""
        
        # Notify all players of creations and deletions
        # NOTE: Normal moves are done above since the SAN moves are calculated before the move...
        # FIXME: Change this so the SAN moves are done afterwards...
        for player in self.__players:
            player.onPieceMoved(piece, start, end)
            
if __name__ == '__main__':
    game = ChessGame()
    
    import pgn
    
    p = pgn.PGN('black.pgn')
    g = p.getGame(0)

    class PGNPlayer(ChessPlayer):
        __moveNumber = 1
        
        __isWhite = True
        
        def __init__(self, isWhite):
            self.__isWhite = isWhite
        
        def readyToMove(self):
            if self.__isWhite:
                move = g.getWhiteMove(self.__moveNumber)
            else:
                move = g.getBlackMove(self.__moveNumber)
            self.__moveNumber += 1
            self.move(move)
            
    white = PGNPlayer(True)
    black = PGNPlayer(False)
    
    game.setWhite(white)
    game.setBlack(black)
    
    game.start()
    