__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

# Extend base UI classes
import sys
sys.path.append('..')
import os

import gobject
import gtk
import gtk.glade
import gtk.gdk
import gtk.gtkgl

import ui

# TEMP
from OpenGL.GL import *

UI_DATA_DIR = 'ui/gtkui/'
if not os.path.isdir(UI_DATA_DIR):
    UI_DATA_DIR = '/usr/share/games/glchess/gui/'

class RemoteGame:
    """
    """
    
    __host = ''
    __serverId = None
    __name = ''
    
    def __init__(self, host, serverId, name):
        self.__host = host
        self.__serverId = serverId
        self.__name = name

    def getHost(self):
        return self.__host
    
    def getServerId(self):
        return self.__serverId
    
    def getGameName(self):
        return self.__name

class GtkServerList:
    __gui = None
    
    __servers = None
    
    def __init__(self, gui):
        self.__gui = gui
        
        self.__servers = []
        view = gui.get_widget('ng_server_list')
        if view is not None:
            store = gtk.ListStore(str, gobject.TYPE_PYOBJECT)
            view.set_model(store)

            cell = gtk.CellRendererText()
            column = gtk.TreeViewColumn('name', cell)
            column.add_attribute(cell, 'text', 0)
            view.append_column(column)
        
    def add(self, host, serverId, name):
        """
        """
        server = RemoteGame(host, serverId, name)
        self.__servers.append(server)
        
        view = self.__gui.get_widget('ng_server_list')
        if view is None:
            return
        model = view.get_model()
        iter = model.append()
        model.set(iter, 0, name)
        model.set(iter, 1, server)
        
    def getSelected(self):
        """
        """
        view = self.__gui.get_widget('ng_server_list')
        if view is None:
            return None
        selection = view.get_selection()
        (model, iter) = selection.get_selected()
        
        if iter is None:
            return None
        else:
            return model.get_value(iter, 1)
        
    def find(self, host, serverId):
        for s in self.__servers:
            if s.getHost() == host and s.getServerId() == serverId:
                return s
        return None
        
    def remove(self, host, serverId):
        """
        """
        s = self.find(host, serverId)
        try:
            self.__servers.remove(s)
        except:
            pass
        
        view = self.__gui.get_widget('ng_server_list')
        if view is None:
            return
        model = view.get_model()
        
        iter = model.get_iter_first()
        while iter is not None:
            if model.get_value(iter, 1) is s:
                break
            iter = model.iter_next(iter)
        
        if iter is not None:
            model.remove(iter)

class GtkNewGameDialog:
    """
    """
    
    # The main UI and the ???
    __mainUI = None
    __gui = None
    
    __serverList = None
    
    __moves = None
    
    def __init__(self, mainUI, aiModel, gameName = None,
                 whiteName = None, blackName = None,
                 whiteAI = None, blackAI = None, moves = None):
        """Constructor for a new game dialog.
        
        'mainUI' is the main UI.
        'aiModel' is the AI models to use.
        'gameName' is the name of the game (string) or None if unknown.
        'whiteName' is the name of the white player (string) or None if unknown.
        'blackName' is the name of the white player (string) or None if unknown.
        'whiteAI' is the type of AI the white player is (string) or None if no AI.
        'blackAI' is the type of AI the black player is (string) or None if no AI.
        'moves' is a list of moves (strings) that the have already been made.
        """
        self.__mainUI = mainUI
        self.__moves = moves
        
        # Load the UI
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'new_game_simple.glade')
        self.__gui.signal_autoconnect(self)
        
        # Make all the AI combo boxes use one list of AI types
        widgets = (self.__getWidget('ng_black_ai_type_combo'),
                   self.__getWidget('ng_white_ai_type_combo'),
                   self.__getWidget('ng_local_ai_type_combo'))
        for widget in widgets:
            if widget is None:
                continue
            
            widget.set_model(aiModel)
            cell = gtk.CellRendererText()
            widget.pack_start(cell, False)
            widget.add_attribute(cell, 'text', 0)
            widget.set_active(0)
            
        haveAIs = aiModel.iter_n_children(None) != 0
        self.__getWidget('white_ai_vbox').set_sensitive(haveAIs)
        self.__getWidget('black_ai_vbox').set_sensitive(haveAIs)
        
        # Use the supplied properties
        if moves:
            self.__getWidget('new_game_dialog').set_title('Restore game (%i moves)' % len(moves))
        if gameName:
            self.__getWidget('ng_game_name_entry').set_text(gameName)
            
        if whiteName:
            self.__getWidget('ng_white_name_entry').set_text(whiteName)
        if blackName:
            self.__getWidget('ng_black_name_entry').set_text(blackName)
            
        # Configure AIs
        if whiteAI:
            self.__getWidget('ng_white_ai_type_combo').set_active_iter(self.__getAIIter(aiModel, whiteAI))
            self.__getWidget('ng_white_ai_radio').set_active(True)
        else:
            self.__getWidget('ng_white_human_radio').set_active(True)
        if blackAI:
            self.__getWidget('ng_black_ai_type_combo').set_active_iter(self.__getAIIter(aiModel, blackAI))
            self.__getWidget('ng_black_ai_radio').set_active(True)
        else:
            self.__getWidget('ng_black_human_radio').set_active(True)
        
        # Show the dialog
        self.__getWidget('new_game_dialog').show()
        self.__testReady()
        
        # ...
        self.__serverList = GtkServerList(self.__gui)
        view = self.__getWidget('ng_server_list')
        if view is not None:
            selection = view.get_selection()
            selection.connect('changed', self.on_ng_properties_changed)

        # TEMP
        self.__serverList.add('alchemy', 1, "Bob's server")

    # Private methods
    
    def __getAIIter(self, model, name):
        """Get an AI engine.
        
        'name' is the name of the AI engine to find.
        
        Return the iter for this AI or None if no AI of this name.
        """
        # FIXME: I'm sure there is a more efficient way of doing this...
        iter = model.get_iter_first()
        while True:
            if name == model.get_value(iter, 0):
                return iter
            
            iter = model.iter_next(iter)
            if iter is None:
                return None

    def __getWidget(self, name):
        """
        """
        return self.__gui.get_widget(name)
    
    def __getAIType(self, comboBox):
        """
        """
        model = comboBox.get_model()
        iter = comboBox.get_active_iter()
        if iter is None:
            return None
        
        data = model.get(iter, 0)
        return data[0]
    
    def __getWhitePlayer(self):
        """
        """
        name = self.__getWidget('ng_white_name_entry').get_text()
        if self.__getWidget('ng_white_human_radio').get_active():
            if len(name) == 0:
                return None
            return (name, None)
        
        elif self.__getWidget('ng_white_ai_radio').get_active():
            if len(name) == 0:
                return None
            aiType = self.__getAIType(self.__getWidget('ng_white_ai_type_combo'))
            if aiType is None:
                return None
            return (name, aiType)
        
        elif self.__getWidget('ng_white_network_radio').get_active():
            return (None, None)
        
    def __getBlackPlayer(self):
        """
        """
        name = self.__getWidget('ng_black_name_entry').get_text()
        if self.__getWidget('ng_black_human_radio').get_active():
            if len(name) == 0:
                return None
            return (name, None)
        
        elif self.__getWidget('ng_black_ai_radio').get_active():
            if len(name) == 0:
                return None
            aiType = self.__getAIType(self.__getWidget('ng_black_ai_type_combo'))
            if aiType is None:
                return None
            return (name, aiType)
        
        elif self.__getWidget('ng_black_network_radio').get_active():
            return (None, None)
  
    def __getLocalPlayer(self):
        """
        """
        name = self.__getWidget('ng_local_name_entry').get_text()
        if len(name) == 0:
            return None

        if self.__getWidget('ng_local_human_radio').get_active():
            return (name, None)
        
        elif self.__getWidget('ng_local_ai_radio').get_active():
            aiType = self.__getAIType(self.__getWidget('ng_local_ai_type_combo'))
            if aiType is None:
                return None
            return (name, aiType)
        
        elif self.__getWidget('ng_local_spectator_radio').get_active():
            return (name, None) # FIXME: Distinguish from human

    def __testReady(self):
        ready = True

        # Depends on what page we start on
        notebook = self.__getWidget('ng_notebook')
        if notebook is None:
            page = 0
        else:
            page = notebook.get_current_page()
        
        # Start game
        if page == 0:
            # Must have a name for the game
            name = self.__getWidget('ng_game_name_entry').get_text()
            if len(name) == 0:
                ready = False
            
            # Must have two valid players
            white = self.__getWhitePlayer()
            black = self.__getBlackPlayer()
            if white is None or black is None:
                ready = False
                
        # Join game
        elif page == 1:
            # Must have a selected server
            if self.__serverList.getSelected() is None:
                ready = False

            # Must have a valid local player
            player = self.__getLocalPlayer()
            if player is None:
                ready = False
                
            # FIXME: Some games do not allow spectators

        else:
            pass #FIXME: raise ABigException

        # Can only click OK if have enough information
        self.__getWidget('ng_ok_button').set_sensitive(ready)

        # Disable properties if toggle buttons not selected
        self.__getWidget('ng_white_ai_properties_table').set_sensitive(self.__getWidget('ng_white_ai_radio').get_active())
        self.__getWidget('ng_black_ai_properties_table').set_sensitive(self.__getWidget('ng_black_ai_radio').get_active())
        table = self.__getWidget('ng_local_ai_properties_table')
        if table is not None:
            table.set_sensitive(self.__getWidget('ng_local_ai_radio').get_active())
        
    def __startGame(self):
        # Depends on what page we start on
        notebook = self.__getWidget('ng_notebook')
        if notebook is None:
            page = 0
        else:
            page = notebook.get_current_page()
        
        # Start game
        if page == 0:        
            # FIXME: Game properties
            gameName = self.__getWidget('ng_game_name_entry').get_text()
            allowSpectators = True

            # Get the players
            white = self.__getWhitePlayer()
            black = self.__getBlackPlayer()
            assert(white is not None)
            assert(black is not None)
            
            # Inform the child class
            self.__mainUI.onGameStart(gameName, allowSpectators, white[0], white[1], black[0], black[1], self.__moves)

        # Join game
        elif page == 1:
            player = self.__getLocalPlayer()
            assert(player is not None)

            # Joining a server
            server = self.__serverList.getSelected()
            assert(server is not None)
            self.__mainUI.onGameJoin(player[0], player[1], server.getHost(), server.getServerId(), server.getGameName())

        else:
            pass #FIXME: raise ABigException
        
    # Gtk+ signal handlers

    def on_ng_find_servers_button_clicked(self, widget, data=None):
        host = self.__getWidget('ng_server_entry').get_text()
        if host == '':
            self.__mainUI.onNetworkServerSearch()
        else:
            self.__mainUI.onNetworkServerSearch(host)

    def on_ng_search_server_entry_changed(self, widget, data=None):
        # FIXME: Change colour back to default
        pass
    
    def on_ng_properties_changed(self, widget, *data):
        self.__testReady()

    def on_response(self, widget, response_id, data = None):
        if response_id == gtk.RESPONSE_OK:
            self.__startGame()
        self.__getWidget('new_game_dialog').destroy()

class GtkLoadGameDialog:
    """
    """
    
    __mainUI = None
    __gui = None
    
    def __init__(self, mainUI):
        """
        """
        self.__mainUI = mainUI
        
        # Load the UI
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'load_game.glade')
        self.__gui.signal_autoconnect(self)
        
        fileChooser = self.__gui.get_widget('filechooserwidget')
        
        # Filter out non PGN files by default
        pgnFilter = gtk.FileFilter()
        pgnFilter.set_name('PGN files')
        pgnFilter.add_pattern('*.pgn')
        fileChooser.add_filter(pgnFilter)
        
        allFilter = gtk.FileFilter()
        allFilter.set_name('All files')
        allFilter.add_pattern('*')
        fileChooser.add_filter(allFilter)
        
        dialog = self.__gui.get_widget('game_load_dialog')
        dialog.show()
        
    def __getFilename(self):
        return self.__gui.get_widget('filechooserwidget').get_filename()
    
    def on_file_changed(self, widget, data = None):
        isFile = os.path.isfile(self.__getFilename())
        self.__gui.get_widget('open_button').set_sensitive(isFile)
        self.__gui.get_widget('properties_button').set_sensitive(isFile)
        
    def on_load_game(self, widget, data = None):
        self.__mainUI.onGameLoad(self.__getFilename(), False)
        self.on_close(widget, data)
    
    def on_configure_game(self, widget, data = None):
        self.__mainUI.onGameLoad(self.__getFilename(), True)
        self.on_close(widget, data)
        
    def on_close(self, widget, data = None):
        self.__gui.get_widget('game_load_dialog').destroy()
        
class GtkSaveGameDialog:
    """
    """
    
    # The main UI
    __mainUI = None

    # The view that is being saved
    __view = None
    
    # The GUI
    __gui = None
    
    def __init__(self, mainUI, view):
        """
        """
        self.__mainUI = mainUI
        self.__view = view
        
        # Load the UI
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'save_game.glade')
        self.__gui.signal_autoconnect(self)
        
        # Filter out non PGN files by default
        dialog = self.__gui.get_widget('dialog')
        pgnFilter = gtk.FileFilter()
        pgnFilter.set_name('PGN files')
        pgnFilter.add_pattern('*.pgn')
        dialog.add_filter(pgnFilter)
        
        allFilter = gtk.FileFilter()
        allFilter.set_name('All files')
        allFilter.add_pattern('*')
        dialog.add_filter(allFilter)
        
    def on_save(self, widget, data = None):
        dialog = self.__gui.get_widget('dialog')
        self.__mainUI._saveView(self.__view, dialog.get_filename())
        dialog.destroy()

    def on_close(self, widget, data = None):
        dialog = self.__gui.get_widget('dialog')
        dialog.destroy()
        self.__mainUI._saveView(self.__view, None)
        
class GtkGameView(gtk.DrawingArea, gtk.gtkgl.Widget):
    """
    """
    # TODO
    __view = None
    
    # TODO
    __glDrawable = None
    
    def __init__(self, glConfig, view):
        """
        """
        gtk.DrawingArea.__init__(self)
        
        self.__view = view
        
        # Allow notification of button presses
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.BUTTON_MOTION_MASK)
        
        # Make openGL drawable
        self.set_gl_capability(glConfig)
        
        # Connect signals
        self.connect('realize', self.__init)
        self.connect('configure_event', self.__configure)
        self.connect('expose_event', self.__expose)
        self.connect('button_press_event', self.__button_press)
        self.connect('button_release_event', self.__button_release)
        
    def redraw(self):
        """
        """
        #FIXME: Check this is valid
        self.window.invalidate_rect(self.allocation, False)
        
    def close(self):
        """
        """
        self.__view.onClose()

    def __startGL(self):
        """
        """
        assert(self.__glDrawable is None)
        
        # Obtain a reference to the OpenGL drawable
        # and rendering context.
        glDrawable = self.get_gl_drawable()
        glContext = self.get_gl_context()

        # OpenGL begin.
        if not glDrawable.gl_begin(glContext):
            return
        
        self.__glDrawable = glDrawable
        
    def __endGL(self):
        """
        """
        assert(self.__glDrawable is not None)
        self.__glDrawable.gl_end()
        self.__glDrawable = None
        
    def __render(self):
        """
        """
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        # Pass onto the user to render the scene
        # Write in Scene class that the user does not need to call glFlush, it will be done internally
        self.__view.onRender()
        
        # Paint this
        if self.__glDrawable.is_double_buffered():
            self.__glDrawable.swap_buffers()
        else:
            glFlush()
        
    def __init(self, widget):
        self.__startGL()
        self.__view.onReshape(widget.allocation.width, widget.allocation.height)
        self.__endGL()
        
    def __configure(self, widget, event):
        self.__startGL()
        self.__view.onReshape(event.width, event.height)
        self.__endGL()
    
    def __expose(self, widget, data = None):
        self.__startGL()
        self.__render()
        self.__endGL()
        
    def __button_press(self, widget, event):
        print 'button_press' + str((event.x, event.y))
        self.__startGL()
        self.__view.onSelect(event.x, event.y)
        self.__endGL()
        
    def __button_release(self, widget, event):
        print 'button_release' + str((event.x, event.y))
        self.__startGL()
        self.__view.onDeselect(event.x, event.y)
        self.__endGL()
        
class GtkGameNotebook(gtk.Notebook):
    """
    """
    
    __glConfig = None
    
    __splashScreen = None
    
    def __init__(self):
        """
        """
        gtk.Notebook.__init__(self)
        
        # Make the tabs scrollable so the area is not resized
        self.set_scrollable(True)
        
        # Configure openGL
        display_mode = (gtk.gdkgl.MODE_RGB | gtk.gdkgl.MODE_DEPTH | gtk.gdkgl.MODE_DOUBLE)
        try:
            self.__glConfig = gtk.gdkgl.Config(mode=display_mode)
        except gtk.gdkgl.NoMatches:
            display_mode &= ~gtk.gdkgl.MODE_DOUBLE
            self.__glConfig = gtk.gdkgl.Config(mode=display_mode)
            
        self.__insertSplashScreen()
        self.set_show_tabs(False)

    def addView(self, name, view):
        """Add a view into the notebook.
        
        'name' is the name of the view.
        'view' is the view containing the callbacks.
        """
        # Get rid of the splashscreen
        if self.__splashScreen is not None:
            self.remove_page(0)
            self.__splashScreen = None
        
        gtkView = GtkGameView(self.__glConfig, view)
        
        page = self.append_page(gtkView)
        self.set_tab_label_text(gtkView, name)
        
        self.set_current_page(page)
        
        self.__updateTabVisibleState()
        
        gtkView.show_all()
        return gtkView
    
    def getView(self, pageNumber = None):
        """Get the view at a given page number.
        
        'pageNumber' is the page to check or None to get the selected page.
        
        Return the view on this page or None if no view here.
        """
        if pageNumber is None:
            # If splashscreen present then there is no view
            if self.__splashScreen is None:
                return self.get_nth_page(self.get_current_page())
            else:
                return None
        else:
            return self.get_nth_page(pageNumber)
    
    def removeView(self, view):
        """Remove a view from the notebook.
        
        'view' is the view to remove.
        """
        self.remove_page(self.page_num(view))
        self.__updateTabVisibleState()
        if self.get_n_pages() == 0:
            self.__insertSplashScreen()

    def __insertSplashScreen(self):
        ui = gtk.glade.XML(UI_DATA_DIR + 'splashscreen.glade', 'about_box')
        self.__splashScreen = ui.get_widget('about_box')
        self.__splashScreen.show()
        self.append_page(self.__splashScreen, gtk.Label('!'))
        
    def __updateTabVisibleState(self):
        """
        """
        # Only show tabs if there is more than one game
        self.set_show_tabs(self.get_n_pages() > 1)

class GtkErrorDialog:
    """
    """
    
    __gui = None
    
    def __init__(self, title, contents):
        """
        """
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'error_dialog.glade')
        self.__gui.signal_autoconnect(self)
        
        self.__gui.get_widget('title_label').set_markup('<b><big>' + title + '</big></b>')
        self.__gui.get_widget('content_label').set_text(contents)
        
        dialog = self.__gui.get_widget('dialog').show_all()

    def on_close(self, widget, data = None):
        dialog = self.__gui.get_widget('dialog').destroy()

class GtkUI(ui.UI):
    """
    """

    # The Gtk+ GUI
    __gui = None

    # The notebook containing games
    __notebook = None
    
    # The Gtk+ list model of the available AIs
    __aiModel = None
    
    # Dictionary of save game dialogs keyed by view
    __saveGameDialogs = None
    
    def __init__(self):
        """Constructor for a GTK+ glChess GUI"""
        self.__saveGameDialogs = {}
        
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'glchess.glade')
        self.__gui.signal_autoconnect(self)

        # Make a notebook for the games
        self.__notebook = GtkGameNotebook()
        self.__notebook.connect('switch-page', self.on_view_changed)
        self.__getWidget('game_viewport').add(self.__notebook)
        self.__notebook.show_all()
        
        # Create the model for the AI types
        self.__aiModel = gtk.ListStore(str)

    # Public methods
    
    def addTimer(self, period, callback):
        """Add a timeout .
        
        'period' is the period for the timer in milliseconds.
        'callback' is the method/function to call when the timer expires.
        
        The timer will be continuously called while 'callback' returns True.
        """
        gobject.timeout_add(period, callback)

    def addAIEngine(self, name):
        """Register an AI engine.
        
        'name' is the name of the engine.
        TODO: difficulty etc etc
        """
        iter = self.__aiModel.append()
        self.__aiModel.set(iter, 0, name)

    def addView(self, name, view):
        """Add a view into the UI.
        
        'name' is the name of the view.
        'view' is the view to add.
        
        Returns a Renderable object.
        """
        return self.__notebook.addView(name, view)
    
    def _saveView(self, view, path):
        """
        """
        self.__saveGameDialogs.pop(view)
        if path is not None:
            self.onGameSave(view, path)
    
    def removeView(self, view):
        """Remove a view from the UI.
        
        'view' is the view to remove.
        """
        self.__notebook.removeView(view)
        self.__updateViewButtons()

    def reportError(self, title, error):
        """Report an error.
        
        'title' is the title of the error (string).
        'error' is the description of the error (string).
        """
        GtkErrorDialog(title, error)
        
    def reportGameLoaded(self, gameName = None,
                         whiteName = None, blackName = None,
                         whiteAI = None, blackAI = None, moves = None):
        """Report a loaded game as required by onGameLoad().
        
        'gameName' is the name of the game (string) or None if unknown.
        'whiteName' is the name of the white player (string) or None if unknown.
        'blackName' is the name of the white player (string) or None if unknown.
        'whiteAI' is the type of AI the white player is (string) or None if no AI.
        'blackAI' is the type of AI the black player is (string) or None if no AI.
        'moves' is a list of moves (strings) that the have already been made.
        """
        GtkNewGameDialog(self, self.__aiModel, gameName = gameName,
                         whiteName = whiteName, whiteAI = whiteAI,
                         blackName = blackName, blackAI = blackAI, moves = moves)

    def run(self):
        """
        """
        gtk.main()

    # Protected methods


    # Private methods

    def __getWidget(self, name):
        return self.__gui.get_widget(name)
    
    def on_view_changed(self, widget, page, pageNum, data = None):
        """Gtk callback"""
        self.__updateViewButtons()
    
    def __updateViewButtons(self):
        """
        """
        view = self.__notebook.getView()
        self.__getWidget('end_game_button').set_sensitive(view is not None)
        self.__getWidget('save_game_button').set_sensitive(view is not None)

    def on_new_game_button_clicked(self, widget, data = None):
        """Gtk callback"""
        GtkNewGameDialog(self, self.__aiModel)
        
    def on_open_game_button_clicked(self, widget, data = None):
        """Gtk callback"""
        GtkLoadGameDialog(self)
        
    def on_save_game_button_clicked(self, widget, data = None):
        """Gtk callback"""
        view = self.__notebook.getView()
        if not self.__saveGameDialogs.has_key(view):
            self.__saveGameDialogs[view] = GtkSaveGameDialog(self, view)
        
    def on_end_game_button_clicked(self, widget, data = None):
        """Gtk callback"""
        view = self.__notebook.getView()
        self.onViewRemoved(view)

    def on_quit(self, widget, data = None):
        """Gtk callback"""
        self.onQuit()

if __name__ == '__main__':
    ui = GtkUI()
    ui.run()
