"""
"""

__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

class Renderable:
    """
    """
    
    def onRender(self):
        pass
    
    def onReshape(self, width, height):
        pass
    
    def onSelect(self, x, y):
        pass
    
    def onDeselect(self, x, y):
        pass
    
    def onClose(self):
        """This method is called when the user closes this view.
        
        TODO
        """
        pass

class UI:
    """Template class for a glChess UI.
    """
    
    # Methods for the glChess to implement
    
    def onGameStart(self, gameName, allowSpectators, whiteName, whiteType, blackName, blackType, moves = None):
        """Called when a local game is started.
        
        'gameName' is the name of the game to create (string).
        'allowSpectators' is a flag to show if remote clients can watch this game (True or False).
        'whiteName' is the name of the white player.
        'whiteType' is the local player type. PLAYER_* or the AI type (string) or None for open.
        'blackName' is the name of the black player.
        'blackType' is the black player type. PLAYER_* or the AI type (string) or None for open.
        'moves' is a list of moves (strings) to start the game with.
        """
        pass
    
    def onGameLoad(self, path, returnResult):
        """Called when a game is loaded.
        
        'path' is the path to the game to load (string).
        'returnResult' is a flag to show if the UI requires the result of the load.
                       If True call reportGameLoaded() if the game can be loaded.
        """
        msg = 'Loading game ' + path
        if configureGame:
            msg += ' after configuring'
        print msg
    
    def onGameJoin(self, localName, localType, serverHost, serverId, game):
        """Called when a network game is started (remote server).
        
        'localName' is the name of the local player (string).
        'localType' is the local player type. PLAYER_* or the AI type (string).
        'serverHost' is the hostname of the server to connect to (string).
        'serverId' is the ID of the server to connect to (int).
        'game' is the game to joing on the server (string).
        """
        print 'Joining game ' + game + ' on server ' + serverHost + ':' + str(serverId) + ' as ' + localName + '(' + str(localType) + ')'
    
    def onNetworkServerSearch(self, hostName=None):
        """Called when the user searches for servers.
        
        'hostName' is the name of the host to look for servers on (string) or
                   None if search whole network.
        """
        pass
    
    def onNetworkGameStart(self, localName, localType, serverHost, serverId):
        """Called when a network game is started (remote server).
        
        'localName' is the name of the local player (string).
        'localType' is the local player type. PLAYER_* or the AI type (string).
        'serverHost' is the hostname of the server to connect to.
        'serverId' is the ID of the server to connect to.
        """
        pass
    
    def onNetworkGameServerStart(self, localName, localType, serverName):
        """Called when a network game is started (local server).
        
        'localName' is the name of the local player (string).
        'localType' is the local player type. PLAYER_* or the AI type (string).
        'serverName' is the name of the server to start.
        """
        pass
    
    def onGameSave(self, view, path):
        """Called when the user requests a game to be saved.
        
        'view' is the game to save (Renderable).
        'path' is the path to the file to save to (string).
        """
        print 'Save game to ' + path
    
    def onViewRemoved(self, view):
        """Called when a user tries to remove a view.
        
        'view' is the view that the user has requested to remove (Renderable).
        """
        pass
    
    def onQuit(self):
        """Called when the user quits the program"""
        pass
    
    # Methods for the UI to implement
    
    def reportError(self, title, error):
        """Report an error.
        
        'title' is the title of the error (string).
        'error' is the description of the error (string).
        """
        pass

    def addView(self, name, view):
        """Add a view into the UI.
        
        'name' is the name of the view.
        'view' is the view to add.
        
        Returns a Renderable object.
        """
        return None
    
    def removeView(self, view):
        """Remove a view from the UI.
        
        'view' is the view to remove (Renderable).
        """
        pass
    
    def reportGameLoaded(self, gameName = None,
                         whiteName = None, blackName = None,
                         whiteAI = None, blackAI = None, moves = None):
        """Report a loaded game as required by onGameLoad().
        
        'gameName' is the name of the game (string) or None if unknown.
        'whiteName' is the name of the white player (string) or None if unknown.
        'blackName' is the name of the white player (string) or None if unknown.
        'whiteAI' is the type of AI the white player is (string) or None if no AI.
        'blackAI' is the type of AI the black player is (string) or None if no AI.
        'moves' is a list of moves (strings) that the have already been made.
        """
        pass
